# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from gboml.compiler.utils import error_
from .objective import Objective
from .constraint import Constraint
from .expression import Expression
from .identifier import Identifier
from .parameter import Parameter
from .link import Attribute
import copy
import numpy as np
from scipy.sparse import coo_matrix  # type: ignore


def is_index_dependant_expression(expression, index) -> bool:

    e_type = expression.get_type()
    nb_child = expression.get_nb_children()
    children = expression.get_children()
    predicate: bool = False
    if e_type == 'literal':

        seed = expression.get_name()
        if type(seed) == Identifier:

            identifier = seed
            id_type = identifier.get_type()
            id_name = identifier.get_name()
            if id_type == "assign":

                predicate = is_index_dependant_expression(
                    identifier.get_expression(), index)
            if id_name in index:

                predicate = True

        elif type(seed) == Attribute:

            attribute = seed
            identifier = attribute.get_attribute()
            if identifier.get_type() == "assign":

                predicate = is_index_dependant_expression(
                    identifier.get_expression(), index)
    else:

        for i in range(nb_child):

            predicate_i = is_index_dependant_expression(children[i], index)
            if predicate_i:

                predicate = predicate_i
                break

    return predicate


class Factorize:

    def __init__(self, obj):

        self.name = None
        type_fact = "sum"
        if type(obj) == Constraint:

            type_fact = "constraint"
        if type(obj) == Objective:

            type_fact = "objective"
        self.type_fact = type_fact
        self.coef_var_tuples = []
        self.indep_expr = None
        self.index_list = []
        self.obj = obj
        self.children = []
        self.mult_expr = None
        self.extension = []
        self.variables = []
        self.line = obj.get_line()
        self.extension_type = ""
        self.sparse = None
        self.independent_terms = []

    def get_name(self):
        return self.name

    def get_line(self):

        return self.line

    def add_coef_var_tuples(self, coef_var):

        self.coef_var_tuples.append(coef_var)

    def get_extension(self):

        return self.extension

    def set_indep_expr(self, expr):

        self.indep_expr = expr
    
    def add_child(self, s):

        self.children.append(s)

    def get_children(self):

        return self.children

    def get_indep_expr(self):

        return self.indep_expr

    def factorize(self, variables, constants, indexes):

        if self.type_fact == "constraint":

            return self.factorize_constraint(variables, constants)
        elif self.type_fact == "objective":

            return self.factorize_objective(variables, constants)
        else:

            return self.factorize_sum(variables, constants, indexes)

    def free(self):
        for child in self.children:
            child.coef_var_tuples = None
            child.obj = None

        self.children = None
        self.coef_var_tuples = None

    def factorize_constraint(self, variables, constants):

        constraint = self.obj
        self.name = constraint.name
        leaves_rhs = constraint.get_rhs().get_leafs()
        leaves_lhs = constraint.get_lhs().get_leafs()
        index_name = constraint.get_index_var() 
        self.index_list.append(index_name)
        var_leaves = []
        term_rhs, _ = self.compute_independant_term(constraint.get_rhs(),
                                                    variables)
        term_lhs, _ = self.compute_independant_term(constraint.get_lhs(),
                                                    variables)
        indep_term = Expression("-")
        indep_term.add_child(term_rhs)
        indep_term.add_child(term_lhs)
        self.set_indep_expr(indep_term)
        for leaf in leaves_rhs:

            inner_expr = leaf.get_name()
            l_type = leaf.get_type()
            if type(inner_expr) == Identifier:

                identifier: Identifier = inner_expr
                identifier_name = identifier.get_name()
                identifier_node_name = identifier.get_node_name()
                if identifier_node_name in variables and \
                        identifier_name in variables[identifier_node_name]:

                    var = variables[identifier_node_name][identifier_name]
                    identifier_var = var.get_identifier()
                    var_leaves.append([-1, leaf, identifier,
                                       identifier_var.get_index(),
                                       identifier_var.get_size()])
                    self.variables.append([identifier_node_name,
                                           identifier_name])

            if type(inner_expr) == Attribute:

                attribute = inner_expr
                attr_node = attribute.get_node_field()
                attr_id = attribute.get_attribute()
                if attr_node in variables:

                    dict_var_node = variables[attr_node]
                    id_name = attr_id.get_name()
                    if id_name in dict_var_node:

                        var = dict_var_node[id_name]
                        var_leaves.append([-1, leaf, attr_id, var.get_index(),
                                           var.get_size()])
                        self.variables.append([attr_node, id_name])
            elif l_type == "sum":

                fct_constr = Factorize(leaf)
                is_var = fct_constr.factorize_sum(variables, constants,
                                                  self.index_list)
                if is_var is True:
                    self.add_child([-1, fct_constr])
                self.variables += fct_constr.variables

        for leaf in leaves_lhs:

            inner_expr = leaf.get_name()
            l_type = leaf.get_type()
            if type(inner_expr) == Identifier:

                identifier: Identifier = inner_expr
                identifier_name = identifier.get_name()
                identifier_node_name = identifier.get_node_name()
                if identifier_node_name in variables and \
                        identifier_name in variables[identifier_node_name]:
                    var = variables[identifier_node_name][identifier_name]
                    identifier_var = var.get_identifier()
                    var_leaves.append([1, leaf, identifier,
                                       identifier_var.get_index(),
                                       identifier_var.get_size()])
                    self.variables.append([identifier_node_name,
                                           identifier_name])
            if type(inner_expr) == Attribute:

                attribute = inner_expr
                attr_node = attribute.get_node_field()
                attr_id = attribute.get_attribute()
                if attr_node in variables:

                    dict_var_node = variables[attr_node]
                    id_name = attr_id.get_name()
                    if id_name in dict_var_node:

                        var = dict_var_node[id_name]
                        var_leaves.append([1, leaf, attr_id, var.get_index(),
                                           var.get_size()])
                        self.variables.append([attr_node, id_name])
            elif l_type == "sum":

                fct_constr = Factorize(leaf)
                is_var = fct_constr.factorize_sum(variables, constants,
                                                  self.index_list)
                if is_var is True:

                    self.add_child([1, fct_constr])
                    self.variables += fct_constr.variables
        coef_var = []
        for rhs_bool, leaf, identifier, index, var_size in var_leaves:

            parent = leaf.get_parent()
            expr = Expression('literal', 1, line=leaf.get_line())
            expr_coef, _ = self.compute_factor(expr, False, parent,
                                               leaf, constants)
            expr = identifier.get_expression()
            coef_var.append([rhs_bool, expr_coef, index, expr, var_size])
        self.coef_var_tuples = coef_var

    def factorize_objective(self, variables, constants):

        objective = self.obj
        self.name = objective.name
        obj_expr = objective.get_expression()
        leaves = obj_expr.get_leafs()
        index_name = objective.get_index_var() 
        self.index_list.append(index_name)
        term_indep, _ = self.compute_independant_term(obj_expr, variables)
        self.set_indep_expr(term_indep)
        var_leaves = []
        for leaf in leaves:

            inner_expr = leaf.get_name()
            l_type = leaf.get_type()
            if type(inner_expr) == Identifier:

                identifier: Identifier = inner_expr
                identifier_node_name = identifier.get_node_name()
                identifier_name = identifier.get_name()
                if identifier_node_name in variables \
                        and identifier_name in variables[identifier_node_name]:

                    var = variables[identifier_node_name][identifier_name]
                    var_identifier = var.get_identifier()
                    var_leaves.append([leaf, var_identifier.get_index(),
                                       var_identifier.get_size()])
            if l_type == "sum":

                fct_constr = Factorize(leaf)
                is_var = fct_constr.factorize_sum(variables, constants,
                                                  self.index_list)
                if is_var is True:

                    self.add_child(fct_constr)
        coef_var = []
        for leaf, index, var_size in var_leaves:

            parent = leaf.get_parent()
            expr = Expression('literal', 1, line=leaf.get_line())
            expr_coef, _ = self.compute_factor(expr, False, parent, leaf,
                                               constants)
            identifier = leaf.get_name()
            expr = identifier.get_expression()
            coef_var.append([expr_coef, index, expr, var_size])
        self.coef_var_tuples = coef_var

    def factorize_sum(self, variables, constants, indexes):

        expression_sum = self.obj
        is_var = False
        var_leaves = []
        children = expression_sum.get_children()
        leaves = children[0].get_leafs()
        time_interval = expression_sum.get_time_interval()
        name_index = time_interval.get_index_name()
        self.index_list = indexes
        self.index_list.append(name_index)
        for leaf in leaves:

            inner_expr = leaf.get_name()
            l_type = leaf.get_type()
            if type(inner_expr) == Identifier:

                identifier: Identifier = inner_expr
                identifier_node_name = identifier.get_node_name()
                identifier_name = identifier.get_name()
                if identifier_node_name in variables and \
                        identifier_name in variables[identifier_node_name]:

                    var = variables[identifier_node_name][identifier_name]
                    variable_identifier = var.get_identifier()
                    var_leaves.append([leaf, variable_identifier.get_index(),
                                       variable_identifier.get_size()])
                    is_var = True
                    self.variables.append(identifier_name)
            elif type(inner_expr) == Attribute:

                attribute = inner_expr
                attr_node = attribute.get_node_field()
                attr_id = attribute.get_attribute()
                if attr_node in variables:

                    dict_var_node = variables[attr_node]
                    id_name = attr_id.get_name()
                    if id_name in dict_var_node:

                        var = dict_var_node[id_name]
                        var_leaves.append([leaf, var.get_index(),
                                           var.get_size()])
                        is_var = True
                        self.variables.append([attr_node, id_name])
            if l_type == "sum":

                fct_constr = Factorize(leaf)
                is_var = fct_constr.factorize_sum(variables,
                                                  constants, self.index_list)
                if is_var is True:

                    self.add_child(fct_constr)
                    is_var = True
                    self.variables += fct_constr.variables
        expr = Expression('literal', 1, line=expression_sum.get_line())
        parent = expression_sum.get_parent()
        expr_coef, _ = self.compute_factor(expr, False, parent,
                                           expression_sum, constants)
        self.mult_expr = expr_coef
        coef_var = []
        for leaf, index, var_size in var_leaves:

            parent = leaf.get_parent()
            expr = Expression('literal', 1, line=leaf.get_line())
            expr_coef, _ = self.compute_factor(expr, False, parent,
                                               leaf, constants,
                                               stop_expr=expression_sum)
            identifier = leaf.get_name()
            if type(identifier) == Attribute:
                identifier = identifier.get_attribute()
            expr = identifier.get_expression()

            coef_var.append([expr_coef, index, expr, var_size])
        self.coef_var_tuples = coef_var

        return is_var

    def compute_independant_term(self, expr, variables):
        
        children = expr.get_children()
        expr_type = expr.get_type()
        is_var = False
        expr_acc = Expression("literal", 0)
        if expr_type == "literal":

            seed = expr.get_name()
            if type(seed) == float or type(seed) == int:

                expr_acc = copy.copy(expr)
            elif type(seed) == Identifier:

                identifier: Identifier = seed
                identifier_name = identifier.get_name()
                identifier_node_name = identifier.get_node_name()
                if identifier_node_name in variables \
                        and identifier_name in variables[identifier_node_name]:

                    is_var = True
                else:

                    expr_acc = copy.copy(expr)
            if type(seed) == Attribute:

                attribute = seed
                attr_node = attribute.get_node_field()
                attr_id = attribute.get_attribute()
                if attr_node in variables:
                
                    dict_var_node = variables[attr_node]
                    id_name = attr_id.get_name()
                    if id_name in dict_var_node:

                        is_var = True
                else:
                
                    expr_acc = copy.copy(expr)
        else:

            tuple_expr_is_var = []
            for child in children:

                expr_child, is_var_child = \
                    self.compute_independant_term(child, variables)
                tuple_expr_is_var.append([expr_child, is_var_child])
            if expr_type in ["*", "/", "**", "mod"]:

                expr1, var_1 = tuple_expr_is_var[0]
                expr2, var_2 = tuple_expr_is_var[1]
                if var_1 or var_2:

                    is_var = True
                else:

                    expr_acc = Expression(expr_type)
                    expr_acc.add_child(expr1)
                    expr_acc.add_child(expr2)

            elif expr_type == "u-":

                expr1, is_var = tuple_expr_is_var[0]
                if is_var is False:

                    expr_acc = Expression("u-")
                    expr_acc.add_child(expr1)
                else:

                    expr_acc = expr1
            elif expr_type in ["+", "-"]:

                expr1, var_1 = tuple_expr_is_var[0]
                expr2, var_2 = tuple_expr_is_var[1]
                if var_1 and var_2:

                    is_var = True
                elif var_1 is False and var_2 is False:

                    expr_acc = Expression(expr_type)
                    expr_acc.add_child(expr1)
                    expr_acc.add_child(expr2)
                elif var_1 is False:

                    expr_acc = expr1
                elif var_2 is False:

                    if expr_type == "-":

                        expr_acc = Expression("u-")
                        expr_acc.add_child(expr2)
                    else:

                        expr_acc = expr2
            elif expr_type == "sum":

                expr1, var_1 = tuple_expr_is_var[0]
                if var_1:

                    is_var = True
                else:

                    time_interval = expr.get_time_interval()
                    expr_acc = Expression('sum')
                    expr_acc.add_child(expr1)
                    expr_acc.set_time_interval(time_interval)

        return expr_acc, is_var

    def compute_factor(self, expr_acc, is_index_dependant, parent_expr,
                       branch_considered, constants, stop_expr=None):
        
        if parent_expr is None or parent_expr == stop_expr or \
                (self.type_fact == "sum" and parent_expr.get_type() == "sum"):

            return expr_acc, is_index_dependant
        children = parent_expr.get_children()
        other_children = []
        p_type = parent_expr.get_type()
        if p_type == "*" or p_type == "/":
            
            i = 0
            for child in children:

                if child != branch_considered:

                    other_children.append((child, i))
                i = i+1
            child, position = other_children[0]
            other_child_time_dependancy = \
                is_index_dependant_expression(child, self.index_list)
            if other_child_time_dependancy is False:

                term1 = child.evaluate_expression(constants)
                if is_index_dependant is False:

                    term2 = expr_acc.evaluate_expression(constants)
                    value = 0
                    if p_type == '*':

                        value = term1 * term2
                    elif p_type == '/':

                        if position == 0:

                            value = term1/term2
                        else:

                            value = term2/term1
                    expr_acc = Expression("literal", value)
                else:

                    copy_child = Expression("literal", term1)
                    expr = Expression(p_type)
                    expr.add_child(expr_acc)
                    expr.add_child(copy_child)
                    expr_acc = expr
            else:

                is_index_dependant = other_child_time_dependancy
                copy_child = copy.copy(child)
                expr = Expression(p_type)
                expr.add_child(expr_acc)
                expr.add_child(copy_child)
                expr_acc = expr
        elif p_type == "-":

            i = 0
            position = 0
            for child in children:

                if child == branch_considered:

                    position = i
                i = i+1
            
            if position == 1:

                if is_index_dependant is False:

                    value = expr_acc.evaluate_expression(constants)
                    expr_acc = Expression("literal", -value)
                else:

                    expr = Expression('u-')
                    expr.add_child(expr_acc)
                    expr_acc = expr
        elif p_type == "u-":

            if is_index_dependant is False:

                value = expr_acc.evaluate_expression(constants)
                expr_acc = Expression("literal", -value)
            else:

                expr = Expression('u-')
                expr.add_child(expr_acc)
                expr_acc = expr
        branch_considered = parent_expr
        parent_expr = parent_expr.get_parent()
        
        return self.compute_factor(expr_acc, is_index_dependant, parent_expr,
                                   branch_considered, constants, stop_expr)

    def extend(self, definitions):

        time_horizon_parameter = definitions["T"]
        time_horizon = time_horizon_parameter.get_value()[0]

        unique_evaluation = False
        coef_var_tuples = self.coef_var_tuples
        nb_coef_var = len(coef_var_tuples)
        explicit_time_range = False
        children_sums = self.get_children()
        np_append = np.append
        list_values_columns = []

        if self.type_fact == "constraint":

            constraint = self.obj
            b_expr = self.get_indep_expr()
            sign = constraint.get_sign()
            self.extension_type = sign
            time_range = constraint.get_time_range(definitions)
            name_index = constraint.get_index_var()
            ignored_times_list = []
            out_of_bounds = []

            if time_range is None:

                t_horizon = time_horizon
                time_range = range(t_horizon)
            else:

                explicit_time_range = True
            if (not is_index_dependant_expression(constraint.get_rhs(),
                                                  name_index)) and \
                    (not is_index_dependant_expression(constraint.get_lhs(),
                                                       name_index)):

                unique_evaluation = True
            all_values = []
            all_columns = []
            all_rows = []
            all_independent_terms = []
            nb_completed_constraints = 0
            for k in time_range:
                index_parameter = Parameter(name_index,
                                            Expression("literal", k))
                index_parameter.set_value([k])
                definitions[name_index] = index_parameter
                flag_out_of_bounds = False
                if constraint.check_time(definitions) is False:
                    ignored_times_list.append(k)
                    continue
                new_values = np.zeros(nb_coef_var)
                columns = np.zeros(nb_coef_var)
                i = 0
                for mult_sign, expr_coef, index, offset_expr, max_size \
                        in coef_var_tuples:

                    new_values[i] = \
                        mult_sign*expr_coef.evaluate_expression(definitions)
                    if offset_expr is not None:

                        offset = offset_expr.evaluate_expression(definitions)
                    else:

                        offset = 0
                    if (offset < 0) or (offset >= max_size):

                        if explicit_time_range:

                            error_("Constraint : "+str(constraint)
                                   + " at line "+str(constraint.get_line())
                                   + " has a time range ill-defined as a "
                                     "variable goes out of bounds for "
                                   + str(name_index) + " equals "+str(k))
                        else:

                            out_of_bounds.append(k)
                        flag_out_of_bounds = True
                        break
                    columns[i] = index + offset
                    i = i+1
                if not flag_out_of_bounds:

                    for sign_mult, child in children_sums:

                        child.extend(definitions)
                        tuple_val_col = child.get_extension()
                        if not tuple_val_col:
                            error_("Out of bounds sum at line "
                                   + str(constraint.get_line()))
                        child_values, child_columns = tuple_val_col
                        child_values = sign_mult * child_values
                        new_values = np_append(new_values, child_values)
                        columns = np_append(columns, child_columns)
                    if b_expr is not None:

                        constant = b_expr.evaluate_expression(definitions)
                    else:

                        constant = 0

                    row_indexes = np.zeros(len(new_values))
                    row_indexes.fill(nb_completed_constraints)
                    all_rows.append(row_indexes)
                    nb_completed_constraints += 1
                    all_values.append(new_values)
                    all_columns.append(columns)
                    all_independent_terms.append(constant)

                    if sign == "==":
                        row_indexes = np.zeros(len(new_values))
                        row_indexes.fill(nb_completed_constraints)
                        all_rows.append(row_indexes)
                        nb_completed_constraints += 1
                        all_values.append(-new_values)
                        all_columns.append(columns)
                        all_independent_terms.append(-constant)

                    # matrix = np.array([new_values, columns])

                    if unique_evaluation is True:

                        break
            if name_index in definitions:

                definitions.pop(name_index)

            if out_of_bounds:

                print("Warning constraint : "
                      + str(constraint) + " at line "
                      + str(constraint.get_line())
                      + " is ignored for "+str(name_index) + " equal to "
                      + str(out_of_bounds))

            if not all_rows:
                self.sparse = None

            else:

                rows = np.concatenate(all_rows)
                columns = np.concatenate(all_columns)
                values = np.concatenate(all_values)
                self.independent_terms = np.array(all_independent_terms,
                                                  dtype=float)
                self.sparse = coo_matrix((values, (rows, columns)),
                                         shape=(nb_completed_constraints,
                                                int(max(columns))+1))
                self.sparse.sum_duplicates()
                self.sparse.eliminate_zeros()

        elif self.type_fact == "objective":

            objective = self.obj
            obj_expr = objective.get_expression()
            obj_range = objective.get_time_range(definitions)
            name_index = objective.get_index_var()
            obj_type = objective.get_type()
            self.extension_type = obj_type
            b_expr = self.get_indep_expr()
            out_of_bounds = []
            ignored_times_list = []

            if obj_range is None:

                t_horizon = time_horizon
                obj_range = range(t_horizon)
            else:

                explicit_time_range = True
            if not is_index_dependant_expression(obj_expr, name_index):

                unique_evaluation = True

            all_values = []
            all_columns = []
            all_rows = []
            all_independent_terms = []
            nb_completed_objectives = 0

            for k in obj_range:

                index_parameter = Parameter(name_index,
                                            Expression("literal", k))
                index_parameter.set_value([k])
                definitions[name_index] = index_parameter
                flag_out_of_bounds = False
                if objective.check_time(definitions) is False:

                    ignored_times_list.append(k)
                    continue
                new_values = np.zeros(nb_coef_var)
                columns = np.zeros(nb_coef_var)
                i = 0
                for expr_coef, index, offset_expr, max_size in coef_var_tuples:

                    new_values[i] = expr_coef.evaluate_expression(definitions)
                    if offset_expr is not None:

                        offset = offset_expr.evaluate_expression(definitions)
                    else:

                        offset = 0
                    if (offset < 0) or (offset >= max_size):

                        if explicit_time_range:

                            error_("Constraint : "+str(objective)
                                   + " at line "+str(objective.get_line())
                                   + " has a time range ill-defined as a "
                                     "variable goes out of bounds for "
                                   + str(name_index) + " equals "+str(k))
                        else:

                            out_of_bounds.append(k)

                        flag_out_of_bounds = True
                        break
                    columns[i] = index + offset
                    i = i+1
                if not flag_out_of_bounds:

                    for child in children_sums:

                        child.extend(definitions)
                        tuple_val_col = child.get_extension()
                        if not tuple_val_col:
                            error_("Out of bounds sum at line "
                                   + str(objective.get_line()))
                        child_values, child_columns = tuple_val_col
                        new_values = np.append(new_values, child_values)
                        columns = np.append(columns, child_columns)

                    constant = b_expr.evaluate_expression(definitions)

                    row_indexes = np.zeros(len(new_values))
                    row_indexes.fill(nb_completed_objectives)
                    all_rows.append(row_indexes)
                    nb_completed_objectives += 1
                    all_values.append(new_values)
                    all_columns.append(columns)
                    all_independent_terms.append(constant)

                    if unique_evaluation is True:

                        break
            if name_index in definitions:

                definitions.pop(name_index)
            if out_of_bounds:

                print("Warning objective : "+str(objective)
                      + " at line "+str(objective.get_line()) +
                      " is ignored for "+str(name_index) + " equal to "
                      + str(out_of_bounds))

            if not all_rows:
                self.sparse = None

            else:

                rows = np.concatenate(all_rows)
                columns = np.concatenate(all_columns)
                values = np.concatenate(all_values)
                self.independent_terms = np.array(all_independent_terms,
                                                  dtype=float)
                self.sparse = coo_matrix((values, (rows, columns)),
                                         shape=(nb_completed_objectives,
                                                int(max(columns))+1))
                self.sparse.sum_duplicates()
                self.sparse.eliminate_zeros()
        elif self.type_fact == "sum":

            expr_sum = self.obj
            time_interval = expr_sum.get_time_interval()
            name_index = time_interval.get_index_name()
            range_index = time_interval.get_range(definitions)
            new_values = np.zeros(nb_coef_var*len(range_index))
            columns = np.zeros(nb_coef_var*len(range_index))
            multiplicator = self.mult_expr.evaluate_expression(definitions)
            flag_out_of_bounds = False
            i = 0
            for k in range_index:

                index_parameter = Parameter(name_index,
                                            Expression("literal", k))
                index_parameter.set_value([k])
                definitions[name_index] = index_parameter
                if expr_sum.check_time(definitions) is False:
                    continue

                for expr_coef, index, offset_expr, max_size in coef_var_tuples:

                    new_values[i] = expr_coef.evaluate_expression(definitions)
                    if offset_expr is not None:

                        offset = offset_expr.evaluate_expression(definitions)
                    else:

                        offset = 0
                    if (offset < 0) or (offset >= max_size):

                        flag_out_of_bounds = True
                        break
                    columns[i] = index + offset
                    i = i+1
                if flag_out_of_bounds:
                    
                    break
            if not flag_out_of_bounds:
                
                for child in children_sums:

                    for k in range_index: 

                        definitions[name_index] = [k]
                        child.extend(definitions)
                        tuple_val_col = child.get_extension()
                        if not tuple_val_col:
                            error_("Out of bounds sum at line "
                                   + str(expr_sum.get_line()))
                        child_values, child_columns = tuple_val_col
                        new_values = np.append(new_values, child_values)
                        columns = np.append(columns, child_columns)
                new_values = new_values*multiplicator
                list_values_columns = [new_values, columns]
            if name_index in definitions:

                definitions.pop(name_index)
        self.extension = list_values_columns



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML lexer file

Defines the tokens of the GBOML language

  Typical usage example:

  tokenize(gboml_file)
  where:
    gboml_file is the file we want to convert to tokens

"""

import ply.lex as lex  # type: ignore
import re


def tokenize(data: str) -> None:
    """tokenize

        takes as input a string and prints it as a stream of tokens

        Args:
            data(str) -> string containing the data to be tokenized

        Returns:

    """

    global lexer
    lexer.input(data)
    while True:
        tok = lexer.token()
        if not tok:
            break
        print(tok)
    lexer = lex.lex()


def tokenize_file(filepath: str) -> None:
    """tokenize_file

        takes as input a filename and prints it as a stream of tokens

        Args:
            filepath(str) -> string containing the filename to print as tokens

        Returns:

    """

    with open(filepath, 'r') as content:
        data = content.read()
    return tokenize(data)


def find_column(input_string, token):
    """
    find_column : input a string and a token
                  find the token column in the string
    """

    line_start = input_string.rfind('\n', 0, token.lexpos) + 1
    return token.lexpos - line_start + 1


keywords = {
    'min': 'MIN',
    'max': 'MAX',
    'external': 'EXTERNAL',
    'internal': 'INTERNAL',
    'integer': 'INTEGER',
    'continuous': 'CONTINUOUS',
    'binary': "BINARY",
    'in': 'IN',
    'import': 'IMPORT',
    'mod': 'MOD',
    'for': 'FOR',
    'where': 'WHERE',
    'with': 'WITH',
    'from': 'FROM',
    'and': 'AND',
    'or': 'OR',
    'not': 'NOT',
    "sum": 'SUM',
    "auxiliary": "AUX",
    "action": "ACTION",
    "state": "STATE",
    "sizing": "SIZING"
    }

reserved = {
    '#NODE': 'NODE',
    '#PARAMETERS': 'PARAM',
    '#CONSTRAINTS': 'CONS',
    '#VARIABLES': 'VAR',
    '#OBJECTIVES': 'OBJ',
    '#TIMEHORIZON': 'TIME',
    "#GLOBAL": "GLOBAL",
    "#HYPEREDGE": "HYPEREDGE",
    "#EXPRESSION": "EXPRESSION",
    }

# List of token names.   This is always required

tokens = (
    'INT',
    'PLUS',
    'MINUS',
    'POW',
    'MULT',
    'DIVIDE',
    'LPAR',
    'RPAR',
    'FLOAT',
    'ID',
    'COMMA',
    'LCBRAC',
    'RCBRAC',
    'LBRAC',
    'RBRAC',
    'EQUAL',
    'LEQ',
    'BEQ',
    'COLON',
    'DOT',
    'FILENAME',
    'SEMICOLON',
    'DOUBLE_EQ',
    'BIGGER',
    'LOWER',
    'NEQ',
    'ASSIGN',
    ) + tuple(keywords.values()) + tuple(reserved.values())

reserved["#LINK"] = "HYPEREDGE"

# Regular expression rules for simple tokens

t_PLUS = r'\+'
t_MINUS = r'-'
t_COMMA = r'\,'
t_LCBRAC = r'\{'
t_RCBRAC = r'\}'
t_LBRAC = r'\['
t_RBRAC = r'\]'
t_POW = r'\*\*'
t_MULT = r'\*'
t_DIVIDE = r'/'
t_LPAR = r'\('
t_RPAR = r'\)'
t_EQUAL = r'\='
t_BIGGER = r'\>'
t_LOWER = r'\<'
t_LEQ = r'\<\='
t_BEQ = r'\>\='
t_COLON = r'\:'
t_SEMICOLON = r'\;'
t_DOT = r'\.'
t_DOUBLE_EQ = r'\=\='
t_NEQ = r'\!\='
t_ASSIGN = r'\<\-'


def t_filename(t):
    r"""\".*\""""

    filename = t.value.replace('"', '')
    legal_characters = r"[a-zA-Z_0-9./:\\]+"
    r = re.compile(legal_characters)

    if re.fullmatch(r, filename):
        t.type = 'FILENAME'
        t.value = filename
        return t
    else:
        t_error(t)


def t_id(t):
    r"""[a-zA-Z_][a-zA-Z_0-9$]*"""

    if t.value in keywords:
        t.type = keywords.get(t.value, 'ID')
    else:
        t.type = "ID"
    return t


def t_reserved(t):
    r"""[#][a-zA-Z_0-9]+"""

    if t.value in reserved:
        t.type = reserved.get(t.value, 'ID')
        return t
    else:
        t_error(t)


def t_comment(t):
    r"""[/][/].*"""

    pass


def t_number(t):
    r"""[0-9]+[\.][0-9]*[e]-[0-9]+|[0-9]+[\.][0-9]*[e][0-9]+|[0-9]+[e][0-9]+|
    [0-9]+[e]-[0-9]+|[0-9]+[\.][0-9]*|[0-9]+|[\.][0-9]+"""

    if "." in t.value or 'e' in t.value:
        t.value = float(t.value)
        t.type = "FLOAT"
    else:
        t.value = int(t.value)
        t.type = "INT"
    return t

# track line number


def t_newline(t):
    r"""\n+"""

    t.lexer.lineno += len(t.value)


# A string containing ignored characters (spaces and tabs)

t_ignore = ' \t\r\f'


# Error handling rule

def t_error(t):

    if lexer.lexdata is not None:
        message = 'Lexing error:' + str(t.lineno) + ':' \
            + str(find_column(lexer.lexdata, t)) + ':'
    else:
        message = 'Lexing error:' + str(t.lineno) + ':'

    if t.type == 'filename':
        message += "Illegal filename '%s'" % t.value
    elif t.type == 'reserved':
        message += "Illegal reserved word '%s'" % t.value
    else:
        message += "Illegal character '%s'" % t.value[0]

    print(message)
    exit(-1)


lexer = lex.lex()



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from gboml.compiler.utils import error_
from .expression import Expression
import sys


class Time:
    """
    Time object is a structure composed of 
    - a variable name 
    - a right handside expression
    - a value (evaluation of the right handside)
    """

    def __init__(self, time_var: str, expr: Expression, line: int = None):

        assert type(time_var) == str, \
            "Internal error: expected string for Time identifier"
        assert type(expr) == Expression, \
            "Internal error: unknown type for expression in Time object"
        if time_var != "T":

            error_("Semantic error:"+str(line)+": Use \"T\"" +
                   " as a symbol for the time horizon. \"" + str(time_var) +
                   "\" is not allowed")
        self.time = time_var
        self.expr = expr
        self.value = expr.evaluate_expression({}) 
        self.line = line

    def __str__(self) -> str:

        string = 'Time Horizon: '+str(self.time)+'\texpr: '+str(self.expr)

        return string

    def get_line(self):
        return self.line

    def get_name(self):

        return self.time

    def set_value(self, value):

        self.value = value

    def get_value(self) -> float:

        return self.value

    def get_expression(self) -> Expression:

        return self.expr

    def check(self) -> None:

        time_value = self.value
        if type(time_value) == float and time_value.is_integer() is False:

            time_value = int(round(time_value))
            print("WARNING: the time horizon considered is not an int")
            print("The time horizon was rounded to "+str(time_value))
        elif type(time_value) == float:

            time_value = int(time_value)
        if time_value < 0:

            error_("ERROR: the chosen time horizon is negative.")
        elif time_value == 0:

            print("WARNING: the time horizon considered is 0")
        self.value = time_value


class TimeInterval:
    """
    Time Interval object is a structure composed of 
    - a variable name 
    - a begin expression
    - an end expression 
    - a step expression or int if not defined
    """

    def __init__(self, time_var: str, begin: Expression,
                 end: Expression, step, line: int):

        assert type(time_var) == str, \
            "Internal error: expected string for TimeInterval identifier"
        assert type(begin) == Expression, \
            "Internal error: unknown type for begin in TimeInterval object"
        assert type(end) == Expression, \
            "Internal error: unknown type for end in TimeInterval object"
        assert type(step) == Expression or type(step) == int, \
            "Internal error: unknown type for step in TimeInterval object"
        self.name = time_var
        if time_var == "t":
            error_("ERROR: t is a reserved keyword "
                   "and can not be overwritten at line "+str(line))

        self.begin = begin
        if type(step) == int:

            self.step = step
        else:

            self.step = step
        self.end = end
        self.line = line

    def __copy__(self):

        time_int = TimeInterval(self.name, self.begin,
                                self.end, self.step, self.line)

        return time_int

    def get_begin(self):

        return self.begin

    def get_step(self):

        return self.step

    def get_end(self):

        return self.end

    def get_index_name(self):

        return self.name
    
    def get_range(self, definitions: dict, clip: int = sys.maxsize) -> range:

        begin_value = self.begin.evaluate_expression(definitions)
        end_value = self.end.evaluate_expression(definitions)
        if type(self.step) == int:

            step_value = self.step
        else:

            step_value = self.step.evaluate_expression(definitions)

        begin_value, end_value, step_value = self.check(begin_value,
                                                        end_value,
                                                        step_value, clip)

        return range(begin_value, end_value+1, step_value)

    def get_interval(self) -> list:

        return [self.begin, self.step, self.end]
    
    def check(self, begin_value, end_value, step_value,
              clip: int = sys.maxsize) -> tuple:

        begin_value = self.convert_type(begin_value, message='begin')
        end_value = self.convert_type(end_value, message="end")
        step_value = self.convert_type(step_value, message="step")
        if end_value < begin_value:

            error_("ERROR: in for loop, the end_value: "+str(self.end) +
                   " is smaller than the begin value "+str(self.begin)
                   + " at line " + str(self.line))

        if step_value < 1:

            error_("ERROR: in for loop, the step value: "+str(self.step)
                   + " is negative or null at line "+str(self.line))
        if end_value+1 > clip:

            print("WARNING: in for loop, end exceeds horizon value " +
                  " end put back to horizon value T at line "+str(self.line))
            end_value = clip

        return begin_value, end_value, step_value
        
    def convert_type(self, value, message: str = "") -> int:
        
        if type(value) == float and value.is_integer() is False:

            value = int(round(value))
            print("WARNING: in for loop, "+message+" value "
                  + " is of type float and was rounded to "+str(value)
                  + " at line "+str(self.line))
        elif type(value) == float:

            value = int(value)
        
        return value



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from setuptools import setup, find_packages
from pathlib import Path

exec(Path('src/gboml/version.py').read_text())

requirements = Path("requirements.txt").read_text().strip().split("\n")

long_description = Path("README.md").read_text()

setup(
    name='gboml',
    description='GBOML: Graph-Based Optimization Modeling Language',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=__version__,
    entry_points={
        'console_scripts': [
            'gboml = gboml:main',
        ]
    },
    author='Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst',
    author_email='bmiftari@uliege.be',
    package_dir={'': 'src'},
    install_requires=requirements,
    extras_require={
        'cplex':  ['cplex'],
        'cbc': ['cylp'],
        'xpress': ['xpress'],
        'gurobi': ['gurobipy'],
        'all_solvers': ['cplex', 'cylp', 'xpress', 'gurobipy']
    },
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Compilers',
    ],
    url='https://gitlab.uliege.be/smart_grids/public/gboml',
    include_package_data=True,
    packages=["gboml"],
)



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML utils file

Defines useful functions that are used throughout the project.

  Typical usage example:

  filename = "/errors/error.txt"
  old_directory, filename = move_to_directory(filename)
  text_file = open(filename, "r")
  error_messages = text_file.read()
  error_(error_messages)

"""
import sys
import os


def check_file_exists(input_file: str):
    if os.path.isfile(input_file) is False:
        print("No such file as "+str(input_file))
        exit(-1)


def move_to_directory(input_file: str):
    """move_to_directory

        takes as input a path to a certain file and moves
        the directory to the one where the file is.

        Args:
            input_file -> string of a file path

        Returns:
            previous directory -> string of the previous directory
            filename -> string of the filename

    """

    check_file_exists(input_file)
    old_directory = os.getcwd()
    directory_path = os.path.dirname(input_file)
    filename = os.path.basename(input_file)

    if directory_path != "":
        os.chdir(directory_path)

    return old_directory, filename


def turn_to_dict(list_nodes: list) -> dict:
    dict_nodes = {}
    for node in list_nodes:
        name = node.get_name()
        if name in dict_nodes:
            error_("The name : "+str(name)
                   + " has already been used at line "
                   + str(dict_nodes[name].get_line())
                   + " redefinition at line "+str(node.get_line()))
        else:
            dict_nodes[name] = node
    return dict_nodes


def flat_nested_list_to_two_level(nested_list_elements):
    all_elements_flat = []
    for node_name, elements in nested_list_elements:
        for element in elements:
            if len(element) != 4 and type(element[0]) != int:
                all_elements_flat += flat_nested_list_to_two_level(element)
            else:
                all_elements_flat.append(element)
    return all_elements_flat


def list_to_string(list_e: list) -> str:
    """list_to_string

        takes as input list of objects and converts them into
        a string of the concatenation

        Args:
            list_e -> list of objects

        Returns:
            string -> string of the concatenation of the string of all objects

    """
    string: str = ""
    for e in list_e:
        string += str(e)+" "
    return string


def update_branch_in_nested_dict(to_update_dict, list_keys,
                                 new_key, to_add_dict):
    if not list_keys:
        to_update_dict[new_key] = to_add_dict
        return [0, to_update_dict]
    else:
        first_key = list_keys.pop(0)
        if first_key in to_update_dict:
            next_level_dict = to_update_dict[first_key]
            code, updated_branch = update_branch_in_nested_dict(next_level_dict,
                                                                list_keys,
                                                                new_key,
                                                                to_add_dict)
            if code == -1:
                return [-1, to_update_dict]

            to_update_dict[first_key] = updated_branch
            return [0, to_update_dict]
        else:
            return [-1, to_update_dict]


def get_branch_in_nested_dict(dictionary, list_keys, not_lower=False):
    accumulator_dict = {}
    level_dict = dictionary
    for key in list_keys:
        if key not in level_dict:
            return [-1, accumulator_dict]

        if key != list_keys[-1] or not not_lower:

            accumulator_dict[key] = level_dict[key]
            level_dict = level_dict[key]
        else:

            nested_dict = level_dict[key]
            intermediate_stopped_dict = dict()
            for nested_key in nested_dict.keys():
                if type(nested_dict[nested_key]) != dict:
                    intermediate_stopped_dict[nested_key] = \
                        nested_dict[nested_key]
            accumulator_dict[key] = intermediate_stopped_dict

    return [0, accumulator_dict]


def get_layer_in_nested_dict(dictionary, list_keys, only_dict=False):
    level_dict = dictionary
    for key in list_keys:
        if key not in level_dict:
            return [-1, level_dict]
        level_dict = level_dict[key]

    if only_dict:
        level_dict_with_only_nested_dicts = {}
        for key in level_dict.keys():
            if type(level_dict[key]) == dict:
                level_dict_with_only_nested_dicts[key] = level_dict[key]
        level_dict = level_dict_with_only_nested_dicts
    return [0, level_dict]


def get_only_objects_in_nested_dict_layer(dictionary):
    only_object_dict = {}
    for key in dictionary.keys():
        if type(dictionary[key]) != dict:
            only_object_dict[key] = dictionary[key]
    return only_object_dict


def error_(message: str) -> None:
    """error_

        takes as input a string message and exits process
        after printing that message

        Args:
            message -> string message to print

    """
    print(message)

    exit(-1)



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""Compiler directory contains each step from the input file to the matrix
generation.

The compiler directory contains :
- The Lexer which converts an input file into a stream of tokens
- The Parser which converts the stream of tokens into a abstract syntax tree
(AST)
- The Semantic Analysis which augments and checks the AST
- The Matrix Generation which converts the AST to a standard optimization form

The file gboml_compiler contains the APIs to compile a full file
The file gboml_lexer contains the APIs to convert a file to a stream of tokens
The file gboml_parser contains the APIs to convert a file to a AST
The file gboml_semantic contains the APIs of the semantic analysis
The file utils contains functions that are used throughout the other files

After running the parser, the PLY package it uses generates two file parser.out
and parsetab.py that are only useful in the parsing step and can be deleted as
they are automatically generated afterwards.
"""


from .gboml_compiler import compile_gboml, compile_gboml_mdp
from .gboml_lexer import tokenize_file
from .gboml_matrix_generation import matrix_generation_a_b, \
    matrix_generation_c, extend_factor, extend_factor_on_multiple_processes
from .gboml_parser import parse_file
from .gboml_semantic import semantic, parameter_evaluation, \
    check_names_repetitions, match_dictionaries, \
    check_mdp, convert_to_mdp, check_program_linearity, factorize_program



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""Cplex Solver file, contains the interface to Cplex solver .

Takes the matrix A, the vector b and vector c as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the cplex solver.

  Typical usage example:

   solution, objective, status, solver_info = cplex_solver(matrix_a, vector_b,
                                                           vector_c,
                                                           objective_offset,
                                                           name_tuples)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""
from gboml.compiler.utils import flat_nested_list_to_two_level

import numpy as np
from scipy.sparse import coo_matrix
import time
import os


def cplex_solver(matrix_a: coo_matrix, vector_b: np.ndarray,
                 vector_c: np.ndarray,
                 objective_offset: float,
                 name_tuples: dict,
                 opt_file: str = None) -> tuple:
    """cplex_solver

        takes as input the matrix A, the vectors b and c. It returns
        the solution of the problem : min c^T * x s.t. A * x <= b found
        by the cplex solver

        Args:
            A -> coo_matrix of constraints
            b -> np.ndarray of independent terms of each constraint
            c -> np.ndarray of objective vector
            objective_offset -> float of the objective offset
            name_tuples -> dictionary of <node_name variables> used
                           to get the type
            opt_file -> optimization parameters file
        Returns:
            solution -> np.ndarray of the flat solution
            objective -> float of the objective value
            status -> solution status
            solver_info -> dictionary of solver information

    """
    try:

        import cplex
    except ImportError:

        print("Warning: Did not find the CPLEX package")
        exit(0)

    if opt_file is None:
        opt_file = 'gboml/solver_api/cplex.opt'
    print(os.getcwd())
    # Convert to appropriate structure
    matrix_a_zipped = zip(matrix_a.row.tolist(), matrix_a.col.tolist(),
                          matrix_a.data)
    m, n = np.shape(matrix_a)
    vector_b = list(vector_b.reshape(-1))
    vector_c = vector_c.tolist()[0]

    # Generate model
    model = cplex.Cplex()
    model.variables.add(obj=vector_c, lb=[-cplex.infinity]*n,
                        ub=[cplex.infinity]*n)
    flat_name_tuples = flat_nested_list_to_two_level(name_tuples)

    for index, _, var_type, var_size in flat_name_tuples:

        if var_type == "integer":

            i = index
            while i < index+var_size:

                model.variables.set_types(i, model.variables.type.integer)
                i = i+1
        if var_type == "binary":

            i = index
            while i < index+var_size:

                model.variables.set_types(i, model.variables.type.binary)
                i = i+1

    model.linear_constraints.add(senses=['L']*m, rhs=vector_b)
    model.linear_constraints.set_coefficients(matrix_a_zipped)
    model.objective.set_sense(model.objective.sense.minimize)
    model.objective.set_offset(objective_offset)

    # Retrieve solver information
    solver_info = {"name": "cplex"}
    print("\nReading CPLEX options from file cplex.opt")
    option_info = {}
    try:

        with open(opt_file, 'r') as optfile:

            lines = optfile.readlines()
    except IOError:

        print("Options file not found")
    else:

        for line in lines:

            line = line.strip()
            option = line.split(" ", 1)
            if option[0] != "":

                try:

                    key = getattr(model.parameters, option[0])
                    assert(isinstance(key, cplex._internal._parameter_classes.Parameter))
                except AttributeError as e:

                    print("Skipping unknown option \'%s\'" % option[0])
                except AssertionError as e:

                    print("Skipping unknown option \'%s\'" % option[0])
                else:

                    if len(option) == 2:

                        try:

                            value = key.type()(option[1])
                        except ValueError as e:

                            print("Skipping option \'%s\' "
                                  "with invalid given value \'%s\' "
                                  "(expected %s)"
                                  % (option[0], option[1], key.type()))
                        else:

                            name = key.__repr__().split(".", 1)[1]
                            print("Setting option \'%s\' to value \'%s\'"
                                  % (name, value))
                            key.set(value)
                            option_info[name] = value
                    else:

                        print("Skipping option \'%s\' with no given value"
                              % option[0])
    solver_info["options"] = option_info
    print("")
    solution = None
    objective = None

    # Solve the problem
    try:

        model.solve()
        status_code = model.solution.get_status()
        solver_info["status"] = status_code
        if status_code == 1 or status_code == 101:

            status = "optimal"
            solution = np.array(model.solution.get_values())
            objective = model.solution.get_objective_value()
        elif status_code == 2 or status_code == 118:

            status = "unbounded"
            objective = float('-inf')
        elif status_code == 3 or status_code == 103:

            status = "infeasible"
            objective = float('inf')
        elif status_code == 23 or status_code == 127:

            status = "feasible"
            solution = np.array(model.solution.get_values())
            objective = model.solution.get_objective_value()
        else:

            status = "unknown"
    except RuntimeError as e:

        print(e)
        status = "error"

    constraints_additional_information = dict()
    variables_additional_information = dict()
    attributes_to_retrieve_constraints = [
                                          ["dual",
                                           model.solution.get_dual_values],
                                          ["slack",
                                           model.solution.get_linear_slacks],
                                          ]
    attributes_to_retrieve_variables = [
                                          ["basis",
                                           model.solution.basis.get_basis, 0],
                                          ["dual_norms",
                                           model.solution.basis.get_dual_norms,
                                           slice(0, 2)]
                                         ]

    for name, function in attributes_to_retrieve_constraints:
        try:
            constraints_additional_information[name] = function()
        except cplex.exceptions.errors.CplexSolverError:
            print("Unable to retrieve ", name, " information for constraints")

    for name, function, index in attributes_to_retrieve_variables:
        try:
            variables_additional_information[name] = function()[index]
        except cplex.exceptions.errors.CplexSolverError:
            print("Unable to retrieve ", name, " information for variables")

    return solution, objective, status, solver_info, \
           constraints_additional_information, \
           variables_additional_information



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Type
from .expression import Expression
from .time_obj import TimeInterval
from .condition import Condition


class Constraint(Type):
    """
    Constraint object is a structure composed of 
    - an operator 
    - a left handside expression 
    - a right handside expression 
    - a TimeInterval object 
    - a condition object
    """

    def __init__(self, c_type: str, lhs: Expression, rhs: Expression,
                 time_interval: TimeInterval = None,
                 condition: Condition = None, line: int = 0):

        assert type(c_type) == str, \
            "Internal error: expected string for Constraint type"
        assert c_type in ["==", "<=", ">="], \
            "Internal error: unknown type for constraint"
        assert type(rhs) == Expression, \
            "Internal error: expected Expression type for " \
            "right hand side in Constraint"
        assert type(rhs) == Expression, "Internal error: expected " \
                                        "Expression type " \
                                        "for left hand side in Constraint"
        assert time_interval is None or type(time_interval) == TimeInterval, \
            "Internal error: expected TimeInterval object in Constraint"
        assert condition is None or type(condition) == Condition, \
            "Internal error: expected Condition object in Constraint"

        Type.__init__(self, c_type, line)
        self.name = None
        self.rhs = rhs
        self.lhs = lhs
        self.time_interval = time_interval
        self.condition = condition

    def __str__(self) -> str:

        string = "[" + str(self.type) + ' , ' + str(self.rhs)
        string += " , " + str(self.lhs)
        if self.condition is not None:
            string += " ][ condition : " + str(self.condition)
        if self.time_interval is not None:
            string += "][ time : " + str(self.time_interval)
        string += ']'

        return string

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    def get_time_interval(self):

        return self.time_interval

    def get_condition(self):

        return self.condition

    def get_sign(self) -> str:

        return self.get_type()

    def get_rhs(self) -> Expression:

        return self.rhs

    def get_lhs(self) -> Expression:

        return self.lhs

    def get_leafs(self) -> list:

        return self.rhs.get_leafs() + self.lhs.get_leafs()

    def get_expanded_leafs(self, dictionary: dict) -> list:

        return self.rhs.expanded_leafs(dictionary) + \
               self.lhs.expanded_leafs(dictionary)

    def get_index_var(self):
        var_name = "t"
        if self.time_interval is not None:
            var_name = self.time_interval.get_index_name()

        return var_name

    def get_time_range(self, definitions):

        range_time = None
        if self.time_interval is not None:
            range_time = self.time_interval.get_range(definitions)

        return range_time

    def check_time(self, definitions):

        predicate = True
        if self.condition is not None:
            predicate = self.condition.check(definitions)

        return predicate
